/**
 * This file contains the methods to compute
 * the distances and times of the cranes among the tanks
 * It also converts the measures from ABS to mm
 */

import java.util.List;
import java.util.HashMap;

public class Calculations {

	private Double[] steps;
	private Double[] speeds;
	private Double ABS;
	private Double ACCELERATION;

	public Calculations(Double ABS_conv, Double acceleration, List<Double> speeds_in_ABS, List<Double> speeds_in_mm)
	{	
	    //List<Object> steps = config.getListOption(Config.SPEED_STEPS_IN_ABS);
		this.steps = speeds_in_ABS.toArray(new Double[] {});
		
		//List<Double> speeds = speeds_in_mm;
		this.speeds = speeds_in_mm.toArray(new Double[] {});

		this.ABS = ABS_conv;
		this.ACCELERATION = acceleration;

	}

	public double[][] calcMatrix(HashMap<Integer,Integer> ABS_tanks, int startTNr, int endTNr)
	{
		int first = startTNr;
		int last = endTNr;

		int nrtanks = last - first + 1;

		double[][] distances = new double[nrtanks][nrtanks];
		double[][] times = new double[nrtanks][nrtanks];


		for(int i = 0; i < nrtanks; i++)
		{
			for(int j = 0; j < nrtanks; j++)
			{

				int dist = ABS_tanks.get(i+first)- ABS_tanks.get(j+first);
				distances[i][j] = Math.abs(dist) * ABS;
			}
		}
		
		
		
		for(int i = 0; i < nrtanks; i++)
		{
			for(int j = 0; j < nrtanks; j++)
			{
				times[i][j] = this.calcTime(distances[i][j]);
			}
		}
	
		return times;
	}
	
	
	
	

	public double calcTimeABS(int from, int to)
	{
		return this.calcTime((to * ABS) - (from * ABS));
	}

	public double calcTimeABS(int distance)
	{
		return this.calcTime(distance * ABS);
	}

	public double calcTime(double from, double to)
	{
		return this.calcTime(to - from);
	}

	public double calcTime(double distance)
	{

		DistanceProfile dp_acc = this.calculateAccelerationProfile(distance);

		DistanceProfile dp_deacc = this.calculateDeaccelerationProfile(distance - dp_acc.distanceTravelled, dp_acc.endSpeed);

		double total_time = dp_acc.timeElapsed + dp_deacc.timeElapsed;

		return total_time;
	}

	/**
	 * Calculate how the crane gets to the destination 
	 * and deaccelerates
	 * 
	 * @param pos_start 
	 * @param speed_start
	 * @return
	 */
	public DistanceProfile calculateDeaccelerationProfile(double pos_start, double speed_start)
	{
		double remaining_distance = pos_start;
		double current_speed = speed_start;
		double time_elapsed = 0;
		
		double step_speed, step_threshold;

		for(int i = steps.length - 1; i >= 0; i--)
		{
			step_speed = (i == steps.length - 1) ? speeds[i] : speeds[i+1];
			step_threshold = i == 0 ? 0 : steps[i];

			/*
			 * If the crane is travelling faster than the allowed 
			 * speed deaccelerate the crane
			 * */
			if(current_speed > step_speed)
			{
				double ttime = this.calcTransitionTime(current_speed, step_speed);

				remaining_distance -= this.calcTransitionDistance(ttime, current_speed, step_speed);
				time_elapsed += ttime;
				current_speed = step_speed;
			}

			/*
			 * If the remaining distance is larger than the current speed 
			 * threshold move the crane to the next threshold at the current speed!
			 * */
			if(remaining_distance > step_threshold*ABS )
			{
				double rest;
				double rest_time;
				/*
				 * If we are at speed step 0 remember to include the time it 
				 * takes to deaccelerate the crane to a halt
				 * 
				 *   /---------                         /
				 *  /            <- like ->    --------/
				 *  
				 * */
				if(i == 0)
				{
					double ttime = this.calcTransitionTime(current_speed, 0);
					remaining_distance -= this.calcTransitionDistance(ttime, current_speed, 0);
					time_elapsed += ttime;
					current_speed = 0;
				} 

				rest = (remaining_distance - step_threshold*ABS);
				rest_time = rest / step_speed;

				time_elapsed += rest_time; 
				remaining_distance = step_threshold*ABS;

			}
		}

		//return new double[] {pos, speed, time};

		DistanceProfile dp = new DistanceProfile();
		dp.distanceTravelled = remaining_distance;
		dp.endSpeed = current_speed;
		dp.timeElapsed = time_elapsed;

		return dp;

	}

	
	/**
	 * Calculate how the crane accelerates and travelles 
	 * before a deacceleration is needed! 
	 * 
	 * @param dist The distance to travel
	 * @return a DistanceProfile describing the distance 
	 * 	travelled, ending speed and elapsed time. 
	 */
	public DistanceProfile calculateAccelerationProfile(double dist)
	{

		double end_acc_after = 0;
		double end_speed = 0;
		double end_time = 0;

		for(int i = this.steps.length - 1; i >= 0; i--)
		{
			double step = steps[i]*ABS;

			/*
			 * If the distance is larger than the current 
			 * deacceleration step try to accelerate and move the crane 
			 * */
			if(dist > step)
			{
				/*
				 * 
				 * */
				double distance_before_deacceleration = dist - step;
				double speed = this.calcFinalSpeed(distance_before_deacceleration);

				if(speed >= this.speeds[i] && speed <= this.speeds[i+1])
				{
					/*
					 * If we during acceleration end up inside a 
					 * deacceleration step? end the acceleration and return
					 * 
					 *     ooooo
					 *     o
					 *     X
					 *     o
					 * ooooo
					 *  
					 *  
					 * */
					end_acc_after = distance_before_deacceleration;
					end_speed = speed;
					break;

				}
				else if(speed < this.speeds[i]) 
				{
					/* continue to next speedstep 
					 * 
					 *     ooooo
					 *     o
					 *     o
					 *     o
					 * ooooo
					 *     
					 *     X
					 *  
					 * */
				}
				else 
				{
					/*
					 * Accelerate the crane to the current speedstep and return
					 * 
					 *     ooooo
					 *     o
					 *     o
					 *     o
					 * ooXoo
					 *  
					 * */

					end_acc_after = this.calcDistByFinalSpeed(this.speeds[i+1]);
					end_speed = this.speeds[i+1];

					break;

				}			
			}

		}
		end_time = end_speed / ACCELERATION;

		//return new double[] {end_acc_after, end_speed,end_time};
		DistanceProfile dp = new DistanceProfile();
		dp.distanceTravelled = end_acc_after;
		dp.endSpeed = end_speed;
		dp.timeElapsed = end_time;

		return dp;
	}

	public double calcFinalSpeed(double dist)
	{
		return Math.sqrt(2 * ACCELERATION * dist);
	}

	public double calcDistByFinalSpeed(double speed)
	{
		return Math.pow(speed, 2) / (2 * ACCELERATION);
	}

	public double calcTransitionTime(double speed_begin, double speed_end)
	{
		return Math.abs((speed_begin - speed_end) / ACCELERATION);
	}
	public double calcTransitionDistance(double time, double speed)
	{
		return time * speed - 0.5 * ACCELERATION * Math.pow(time,2);
	}
	public double calcTransitionDistance(double time, double speed_start, double speed_end)
	{
		return 0.5 * (speed_start + speed_end) * time;
	}


	private class DistanceProfile 
	{
		public double endSpeed;
		public double distanceTravelled;
		public double timeElapsed;

	}

}

